// Generator for the video Dreamscape: The Missing Scene: https://www.youtube.com/watch?v=pgRTLwWNOSY
// Written by Nils Liaaen Corneliusen 2020.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// https://www.ignorantus.com

// ffmpeg -framerate 60 -i frame-%05d.jpg -framerate 60 -c:v libx264 -preset fast -crf 18 -pix_fmt yuv420p -s 1920x1080 test5.mp4
// ffmpeg -framerate 60 -i frame-%05d.jpg -i C:\Users\ncorn\Recycler\Assets\Werq_32.mp3 -framerate 60 -c:v libx264 -preset slow -crf 18 -pix_fmt yuv420p -s 1920x1080 -movflags +faststart test6.mp4
// ffmpeg -framerate 60 -i frame-%05d.jpg -i C:\Users\ncorn\Recycler\Assets\supernatural.mp3 -framerate 60 -c:v libx264 -preset slow -crf 18 -pix_fmt yuv420p -s 1920x1080 -movflags +faststart test10.mp4
#include <iostream>
#include <Windows.h>
#include <jpeglib.h>

#include "timer.h"
#include "scale.h"
#include "coeffs.h"
#include "bmp.h"

#define WIDTH 1920
#define HEIGHT 1080

#define FRAMES_PATH "C:\\Users\\ncorn\\Recycler\\frames\\"
#define ASSETS_PATH "C:\\Users\\ncorn\\Recycler\\assets\\"

#define ATOM_FILE ASSETS_PATH "atom.yuv"
#define ATOM_FRAMES_IN_FILE 250
#define ATOM_FRAMES 248 // skip 2 first, first and last frame black


uint8_t *loadfile( const char *fname, size_t *sz )
{
	*sz = 0;

	FILE *fp = fopen( fname, "rb" );
	if( fp == NULL ) {
		printf( "could not open file %s\n", fname );
		return NULL;
	}

	fseek( fp, 0L, SEEK_END );
	size_t size = ftell( fp );
	fseek( fp, 0L, SEEK_SET );

	if( size == 0 ) {
		printf( "%s has size 0\n", fname );
		return NULL;
	}

	*sz = size;

	uint8_t *buf = (uint8_t *)malloc( size );

	size_t rc = fread( buf, 1, size, fp );
	fclose( fp );
	if( rc != size ) {
		printf( "%s: short read. got %zu, expected %zu\n", fname, rc, size );
		free( buf );
		return NULL;
	}

//	printf( "loaded %s:  %zu bytes\n", fname, size );

	return buf;
}

void saveframe( bmp_argb *src, int framectr )
{
	char fname[160];
	sprintf( fname, "%sframe-%05d.jpg", FRAMES_PATH, framectr );
	bmp_save_jpg( src, fname, false, 100 );
}

#define MIN(a,b) ((a)<(b)?(a):(b))

bool saveJPEG(uint8_t *src, int framectr, int quality)
{
	char fname[160];
	sprintf( fname, "%sframe-%05d.jpg", FRAMES_PATH, framectr );

	printf( "save %s\n", fname );

	FILE *fp = fopen( fname, "wb" );
    if( fp == NULL ) {
        printf( "Could not create ouput file %s\n", fname );
        return false;
    }

	// More info on YUV420P encoding -> 3.11 Raw (downsampled) image data
	// https://www4.cs.fau.de/Services/Doc/graphics/doc/jpeg/libjpeg.html
	// https://stackoverflow.com/questions/16390783/how-to-compress-yuyv-raw-data-to-jpeg-using-libjpeg

	jpeg_compress_struct cinfo = {0};
	jpeg_error_mgr jerr = {0};

	cinfo.err = jpeg_std_error(&jerr);

    jpeg_create_compress( &cinfo );

    jpeg_stdio_dest(&cinfo, fp);

	cinfo.image_width = 1920; /* Four fields of the cinfo struct must be filled in: */
	cinfo.image_height = 1080;

	cinfo.in_color_space = JCS_YCbCr;
	cinfo.input_components = 3;

	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE);

	// yuv444
    cinfo.comp_info[0].v_samp_factor = 1;
    cinfo.comp_info[0].h_samp_factor = 1;

	cinfo.raw_data_in = true;
	jpeg_start_compress(&cinfo, true);

	uint8_t *yplane = src;
	uint8_t *uplane = src + 1920*1080;
	uint8_t *vplane = src + 1920*1080*2;

	JSAMPROW y_row[16];
	JSAMPROW u_row[16];
	JSAMPROW v_row[16];

	JSAMPARRAY rows[]  { y_row, u_row, v_row };

	while(cinfo.next_scanline < cinfo.image_height) {

		for( int i = 0; i < 16; i++ ) {
			y_row[i] = yplane + 1920*MIN( cinfo.next_scanline+i, cinfo.image_height-1 );
			u_row[i] = uplane + 1920*MIN( cinfo.next_scanline+i, cinfo.image_height-1 );
			v_row[i] = vplane + 1920*MIN( cinfo.next_scanline+i, cinfo.image_height-1 );
		}

		int rowproc = jpeg_write_raw_data(&cinfo, rows, 16);
	}

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

	fclose( fp );

	return true;
}

void resample( uint8_t *src, int srcw, int srch, int srcstr )
{
	for( int y = 0; y < srch; y++ ) {
	
		uint8_t *row = src + srcw*y;

		for( int x = 0; x < srcw; x++ ) {
			double b = *row;
			
			if( b < 16.0 ) b = 16.0;
			if( b > 240.0 ) b = 240.0;

			b -= 16.0;
			b *= 1.138392857142857;
			b = round( b );
			if( b > 255.0 ) {
				printf( "eeeek %f\n", b );
				exit( 1 );
			}
			uint8_t ub = (uint8_t)b;

			*row++ = ub;
		}
	
	}

}

#define ANIM_W 160
#define ANIM_H 120

static void tempify( uint8_t *src0, uint8_t *src1, uint8_t *dst, double bw )
{
	double aw = 1.0 - bw;

	for( int y = 0; y < ANIM_H; y++ ) {

		uint8_t *src0y = src0  + ANIM_W*y;
		uint8_t *src0u = src0y + ANIM_W*ANIM_H;
		uint8_t *src0v = src0u + ANIM_W*ANIM_H;

		uint8_t *src1y = src1  + ANIM_W*y;
		uint8_t *src1u = src1y + ANIM_W*ANIM_H;
		uint8_t *src1v = src1u + ANIM_W*ANIM_H;

		uint8_t *dsty = dst  + ANIM_W*y;
		uint8_t *dstu = dsty + ANIM_W*ANIM_H;
		uint8_t *dstv = dstu + ANIM_W*ANIM_H;

		for( int x = 0; x < ANIM_W; x++ ) {
			double s0y = *src0y++ * aw;
			double s1y = *src1y++ * bw;
			int outy = (int)(round(s0y + s1y));

			double s0u = *src0u++ * aw;
			double s1u = *src1u++ * bw;
			int outu = (int)(round(s0u + s1u));

			double s0v = *src0v++ * aw;
			double s1v = *src1v++ * bw;
			int outv = (int)(round(s0v + s1v));

			*dsty++ = outy;
			*dstu++ = outu;
			*dstv++ = outv;
		}
		
	}

}

uint8_t *atom_base;
uint8_t *atom;

bool prep_atom( void )
{
	size_t fsize;

	atom_base = loadfile( ATOM_FILE, &fsize );
	if( atom_base == NULL ) return false;

	for( int i = 0; i < ATOM_FRAMES_IN_FILE; i++ ) {
		resample( atom_base + 160*120*3*i, 160, 120, 160 );
	}

	atom = atom_base + 160*120*3*2;
	memset( atom+160*120*3*247, 0x00, 160*120*1 );
	memset( atom+160*120*3*247+160*120, 0x80, 160*120*2 );

	return true;
}

int gen_atom( int framectr, bool rev )
{
	uint8_t *scaleframe = (uint8_t *)malloc( WIDTH*HEIGHT*3 );
	memset( scaleframe, 0x00, WIDTH*HEIGHT );
	memset( scaleframe + WIDTH*HEIGHT*1, 0x80, WIDTH*HEIGHT*2 );

//	float scalefac = 0.65f;
	float scalefac = 0.80f;

	uint32_t cox[COEFFS_PHASES];
	uint32_t coy[COEFFS_PHASES];

	coeffs_get( scalefac, cox );
	coeffs_get( scalefac, coy );

	int startframe = rev ? ATOM_FRAMES-1 : 0;
	int endframe   = rev ? 0 : ATOM_FRAMES-1;
	int adder = rev ? -1 : 1;

	for( int i = startframe; i != endframe+adder; i += adder ) {
		uint8_t *tmpbuf = (uint8_t *)malloc( 160*120*3 );
		uint8_t *src0 = atom + 160*120*3*i;
		uint8_t *src1;
		if( rev ) {
			src1 = i > 0 ? src0-160*120*3 : src0;
		} else {
			src1 = i < ATOM_FRAMES-1 ? src0+160*120*3 : src0;
		}

		scale_plane( src0 + 160*120*0, 160, 120, 160, scaleframe+1920*1080*0+240, 1440, 1080, 1920, cox, coy );
		scale_plane( src0 + 160*120*1, 160, 120, 160, scaleframe+1920*1080*1+240, 1440, 1080, 1920, cox, coy );
		scale_plane( src0 + 160*120*2, 160, 120, 160, scaleframe+1920*1080*2+240, 1440, 1080, 1920, cox, coy );

		saveJPEG( scaleframe, framectr, 100 );
		framectr++;

		for( int j = 1; j < 5; j++ ) {

			double bw = 0.20 * j;

			tempify( src0, src1, tmpbuf, bw );

			scale_plane( tmpbuf + 160*120*0, 160, 120, 160, scaleframe+1920*1080*0+240, 1440, 1080, 1920, cox, coy );
			scale_plane( tmpbuf + 160*120*1, 160, 120, 160, scaleframe+1920*1080*1+240, 1440, 1080, 1920, cox, coy );
			scale_plane( tmpbuf + 160*120*2, 160, 120, 160, scaleframe+1920*1080*2+240, 1440, 1080, 1920, cox, coy );

			saveJPEG( scaleframe, framectr, 100 );
			framectr++;

		}

	}
	free( scaleframe );

	return framectr;
}

int gen_black( int framectr, int cnt )
{
	bmp_argb *black = bmp_alloc( 1920, 1080 );
	for( int j = 0; j < cnt; j++ ) {
		saveframe( black, framectr );
		framectr++;
	}
	bmp_free( black );
	return framectr;
}

int gen_pic( int framectr, bmp_argb *src, int cnt )
{
	for( int j = 0; j < cnt; j++ ) {
		saveframe( src, framectr );
		framectr++;
	}
	return framectr;
}

int main( int argc, char *argv[] )
{
	int framectr = 0;

	bool ok = prep_atom();
	if( !ok ) return 1;

	bmp_argb *trm = bmp_load( ASSETS_PATH "triumph_1080.bmp", true );
	if( trm == NULL ) return 1;

	bmp_argb *cj = bmp_load( ASSETS_PATH "triumph_1080_end.bmp", true );
	if( cj == NULL ) return 1;

	bmp_argb *middle = bmp_load( ASSETS_PATH "triumph_mp_1080.bmp", true );
	if( middle == NULL ) return 1;

	bmp_argb *ufc = bmp_load( ASSETS_PATH "ufc.bmp", true );
	if( ufc == NULL ) return 1;

	framectr = gen_pic( framectr, trm, 240 );

	framectr = gen_atom( framectr, false );

	framectr = gen_pic( framectr, middle, 240 );

	framectr = gen_atom( framectr, true );
	framectr = gen_pic( framectr, cj, 300 );
	printf( "done: %d frames\n", framectr );



	printf( "Goodbye, Mr Bond\n" );

	return 0;
}
